package Orangutan::Context;

use strict;

use Data::Dumper;

sub new {
    my $class = shift;

    my $self = { };
    bless $self, $class;

    if (ref $_[0] eq 'Orangutan::Context') {
        my $context = shift;
        foreach my $member (keys %{$context}) {
            $self->{$member} = $context->{$member};
        }
        if (defined($self->{'generic'})) {
            delete($self->{'generic'});
        }
    } else {
        my %args = ( );
        while ($#_ >= 0) {
            $args{lc(pop)} = pop;
        }
        foreach my $arg (keys %args) {
            # Initialization
            if ($arg eq 'init') {
                $self->{'init'} = $args{$arg};
            # Question/sentence for bot to send
            } elsif (($arg eq 'question') ||
                ($arg eq 'request')) {
                $self->{'out'} = $args{$arg};
            # Delay for request
            } elsif ($arg eq 'delay') {
                $self->{'delay'} = $args{$arg};
            # Answer to get
            } elsif
               (($arg eq 'answer') ||
                ($arg eq 'response')) {
                $self->{'in'} = $args{$arg};
            # Reply to answer or request from user
            } elsif ($arg eq 'reply') {
                $self->{'reply'} = $args{$arg};
            # Timeout for answer
            } elsif
               (($arg eq 'time') ||
                ($arg eq 'timeout')) {
                $self->{'time'} = $args{$arg};
            # Response handler
            } elsif
               (($arg eq 'handler') ||
                ($arg eq 'onmatch')) {
                $self->{'handler'} = $args{$arg};
            # Handler called on timeout
            } elsif ($arg eq 'ontimeout') {
                $self->{'timedout'} = $args{$arg};
            # Called if there was no match
            } elsif ($arg eq 'nomatch') {
                $self->{'nomatch'} = $args{$arg};
            # Event handler
            } elsif ($arg eq 'event') {
                $self->{'event'} = $args{$arg};
            # Undo handler
            } elsif ($arg eq 'undo') {
                $self->{'undo'} = $args{$arg};
            # Scheduler
            } elsif ($arg eq 'schedule') {
                $self->{'schedule'} = $args{$arg};
            # Help
            } elsif ($arg eq 'help') {
                my %help = ( );
                foreach my $key (keys %{$args{$arg}}) {
                    if (($key eq 'title') ||
                        ($key eq 'question') ||
                        ($key eq 'answer') ||
                        ($key eq 'weight')) {
                        $help{$key} = $args{$arg}->{$key};
                    } else {
                        $main::logger->Log("unsupported help property: %s", $key);
                    }
                }
                if (defined($help{'title'}) &&
                    defined($help{'question'}) &&
                    defined($help{'answer'})) {
                    $self->{'help'} = \%help;
                } else {
                    $main::logger->Log("ignoring incomplete help data");
                }
            # Weight
            } elsif ($arg eq 'weight') {
                if (!defined($args{$arg})) {
                    $self->{'weight'} = [ undef, undef ];
                } elsif (ref $args{$arg} eq 'HASH') {
                    $self->{'weight'} = [ undef, undef ];
                    if (defined($args{$arg}->{'out'})) {
                        $self->{'weight'}[0] = $args{$arg}->{'out'};
                    }
                    if (defined($args{$arg}->{'request'})) {
                        $self->{'weight'}[0] = $args{$arg}->{'request'};
                    }
                    if (defined($args{$arg}->{'in'})) {
                        $self->{'weight'}[1] = $args{$arg}->{'in'};
                    }
                    if (defined($args{$arg}->{'response'})) {
                        $self->{'weight'}[1] = $args{$arg}->{'response'};
                    }
                } elsif
                   ((ref $args{$arg} eq 'ARRAY') ||
                    (ref $args{$arg} eq 'CODE')) {
                    $self->{'weight'} = $args{$arg};
                } else {
                    $main::logger->Log("invalid weight for context");
                }
            # Flag = indicates that handlers write to custom fields f.e.
            } elsif ($arg eq 'copy') {
                $self->{'copy'} = $args{$arg};
            } else {
                $main::logger->Log("unsupported context property: %s", $arg);
            }
        }
        if (defined($self->{'out'}) &&
            defined($self->{'in'})) {
            $self->{'inactive'} = 1;
        }
        $self->{'generic'} = 1;
    }

    return $self;
}

sub SetID($$) {
    my ($self, $id) = @_;
    $self->{'id'} = $id;
}

sub SetField($$$) {
    my ($self, $field, $value) = @_;
    if (!defined($self->{'custom'})) {
        $self->{'custom'} = { };
    }
    $self->{'custom'}->{$field} = $value;

    if (defined($self->{'generic'})) {
        $main::logger->Log("generic context %s sets field", $self->GetID);
    }
}

sub SetTime($;$) {
    my ($self, $time) = @_;
    if (defined($time)) {
        $self->{'time'} = $time;
    } else {
        delete($self->{'time'});
    }
}

sub SetTimeout($$) {
    my ($self, $timeout) = @_;
    if (defined($timeout)) {
        $self->{'timeout'} = $timeout;
    } else {
        delete($self->{'timeout'});
    }
}

sub SetDelay($$) {
    my ($self, $delay) = @_;
    $self->{'delay'} = $delay;
}

sub UnsetField($$) {
    my ($self, $field) = @_;
    if (defined($self->{'custom'}) && defined($self->{'custom'}->{$field})) {
        delete($self->{'custom'}->{$field});
    }
}

sub InsertForeignHandler($$;$) {
    my ($self, $foreign, $timeout) = @_;
    if (!$timeout) {
        $timeout = 1800;
    }
    $timeout += time;
    if (defined($self->{'handlers'})) {
        for (my $i = 0; $i < scalar @{$self->{'handlers'}}; $i++) {
            if ($self->{'handlers'}[$i][1] eq $foreign) {
                splice(@{$self->{'handlers'}}, $i, 1);
                last;
            }
        }
        unshift(@{$self->{'handlers'}}, [ $timeout, $foreign ]);
    } else {
        $self->{'handlers'} = [ [ $timeout, $foreign ] ];
    }

    if (defined($self->{'generic'})) {
        $main::logger->Log("generic context %s inserts foreign handler", $self->GetID);
    }
}

sub RemoveForeignHandler($$) {
    my ($self, $foreign) = @_;
    if (defined($self->{'handlers'})) {
        for (my $i = 0; $i < scalar @{$self->{'handlers'}}; $i++) {
            if ($self->{'handlers'}[$i][1] eq $foreign) {
                splice(@{$self->{'handlers'}}, $i, 1);
                last;
            }
        }
    }
}

sub CleanForeignHandlers($) {
    my $self = shift;
    if (defined($self->{'handlers'})) {
        my $time = time;
        for (my $i = $#{$self->{'handlers'}}; $i >= 0; $i--) {
            if (($self->{'handlers'}[$i][0] > 0) && ($self->{'handlers'}[$i][0] < $time)) {
                splice(@{$self->{'handlers'}}, $i, 1);
            }
        }
    }
}

sub SuspendForeignHandlers($) {
    my $self = shift;
    if (defined($self->{'handlers'})) {
        my $time = time;
        for (my $i = 0; $i < scalar @{$self->{'handlers'}}; $i++) {
            if ($self->{'handlers'}[$i][0] > $time) {
                $self->{'handlers'}[$i][0] = $time - $self->{'handlers'}[$i][0];
            }
        }
    }
}

sub ResumeForeignHandlers($) {
    my $self = shift;
    if (defined($self->{'handlers'})) {
        my $time = time;
        for (my $i = 0; $i < scalar @{$self->{'handlers'}}; $i++) {
            if ($self->{'handlers'}[$i][0] < 0) {
                $self->{'handlers'}[$i][0] = $time - $self->{'handlers'}[$i][0];
            }
        }
    }
}

sub ShouldCopy($) {
    my $self = shift;
    if (defined($self->{'generic'}) &&                          # only a generic context should be copied
        ((defined($self->{'out'}) && defined($self->{'in'})) || # will save active state
        defined($self->{'copy'}) ||                             # some other reasons to copy (e.g. fields)
        $self->ShouldWait ||                                    # has a timeout (saves timeout info)
        $self->HasDelay)) {                                     # has a delay (saves delay info)
        return 1;
    } else {
        return 0;
    }
}

sub ShouldTell($) {
    my $self = shift;
    if (defined($self->{'out'})) {
        return 1;
    } else {
        return 0;
    }
}

sub ShouldWait($) {
    my $self = shift;
    if (defined($self->{'time'}) && ($self->{'time'} > 0)) {
        return 1;
    } else {
        return 0;
    }
}

sub ShouldReply($) {
    my $self = shift;
    if (defined($self->{'reply'})) {
        return 1;
    } else {
        return 0;
    }
}

sub IsActive($) {
    my $self = shift;
    return !defined($self->{'inactive'});
}

sub IsReusable($) {
    my $self = shift;
    return !defined($self->{'time'});
}

sub IsGeneric($) {
    my $self = shift;
    return defined($self->{'generic'});
}

sub HasInit($) {
    my $self = shift;
    if (defined($self->{'init'})) {
        return 1;
    } else {
        return 0;
    }
}

sub HasHandler($) {
    my $self = shift;
    if (defined($self->{'handler'})) {
        return 1;
    } else {
        return 0;
    }
}

sub HasTimeoutHandler($) {
    my $self = shift;
    if (defined($self->{'timedout'})) {
        return 1;
    } else {
        return 0;
    }
}

sub HasNoMatchHandler($) {
    my $self = shift;
    if (defined($self->{'nomatch'})) {
        return 1;
    } else {
        return 0;
    }
}

sub HasEventHandler($) {
    my $self = shift;
    if (defined($self->{'event'})) {
        return 1;
    } else {
        return 0;
    }
}

sub HasDelay($) {
    my $self = shift;
    if (defined($self->{'delay'})) {
        return 1;
    } else {
        return 0;
    }
}

sub HasSchedule($) {
    my $self = shift;
    if (defined($self->{'schedule'})) {
        return 1;
    } else {
        return 0;
    }
}

sub HasUndo($) {
    my $self = shift;
    if (defined($self->{'undo'})) {
        return 1;
    } else {
        return 0;
    }
}

sub HasForeignHandlers($) {
    my $self = shift;
    return defined($self->{'handlers'}) && (scalar @{$self->{'handlers'}} > 0);
}

sub InitTimeout($) {
    my $self = shift;
    $self->{'timeout'} = $self->{'time'} + time;
}

sub InitDelay($) {
    my $self = shift;
    $self->{'delay'} = $self->{'delay'} + time;
}

sub Activate($) {
    my $self = shift;
    if (defined($self->{'inactive'})) {
        delete($self->{'inactive'});
    }
}

sub CallInit(@) {
    return $_[0]->{'init'}(@_);
}

sub CallHandler(@) {
    $_[1]->UpdateStack($_[0]);
    return $_[0]->{'handler'}(@_);
}

sub CallTimeoutHandler(@) {
    return $_[0]->{'timedout'}(@_);
}

sub CallNoMatchHandler(@) {
    return $_[0]->{'nomatch'}(@_);
}

sub CallEventHandler(@) {
    if (scalar @_ >= 3) {
        return $_[0]->{'event'}->{$_[2]}(@_);
    } else {
        return undef;
    }
}

sub CallForeignHandler($) {
    my $self = shift;
    if (defined($self->{'handlers'})) {
        my $time = time;
        for (my $i = $#{$self->{'handlers'}}; $i >= 0; $i--) {
            if ($self->{'handlers'}[$i][0] < $time) {
                splice(@{$self->{'handlers'}}, $i, 1);
            }
        }
        if (scalar @{$self->{'handlers'}} > 0) {
            my $context = $self->{'handlers'}[0][1];
            my $keep = $context->CallHandler(@_);
            if (!$context->IsReusable && !$keep) {
                $_[0]->RemoveResponse($context);
            }
            return 1;
        }
    }
    return 0;
}

sub CallSchedule(@) {
    return $_[0]->{'schedule'}(@_);
}

sub CallUndo() {
    return $_[0]->{'undo'}(@_);
}

sub GetID($) {
    my $self = shift;
    return $self->{'id'};
}

sub GetWeight($) {
    my $self = shift;
    my ($out, $in) = (undef, undef);
    if (defined($self->{'weight'})) {
        if (ref $self->{'weight'} eq 'CODE') {
            ($out, $in) = $self->{'weight'}($self, @_);
        } else {
            ($out, $in) = @{$self->{'weight'}};
        }
        if (!defined($self->{'out'}) && defined($out)) {
            $main::logger->Log("context %s set for output but no output defined", $self->GetID);
            $out = undef;
        }
        if (!defined($self->{'in'}) && defined($in)) {
            $main::logger->Log("context %s set for input but no input defined", $self->GetID);
            $out = undef;
        }
    } else {
        if (defined($self->{'out'})) {
            $out = 0;
        }
        if (defined($self->{'in'})) {
            $in = 0;
        }
    }
    return ($out, $in);
}

sub GetField($$) {
    my ($self, $field) = @_;
    return (defined($self->{'custom'}) && defined($self->{'custom'}->{$field})) ? $self->{'custom'}->{$field} : undef;
}

sub GetTimeout($) {
    my $self = shift;
    return (defined($self->{'timeout'})) ? $self->{'timeout'} : undef;
}

sub GetDelay($) {
    my $self = shift;
    return (defined($self->{'delay'})) ? $self->{'delay'} : undef;
}

sub GetHelp($) {
    my $self = shift;
    return $self->{'help'};
}

sub Tell($$) {
    my ($self, $user) = @_;
    my $body = undef;
    if (ref $self->{'out'} eq 'ARRAY') {
        $body = $self->Random($self->{'out'});
    } elsif (ref $self->{'out'} eq 'CODE') {
        $body = $self->{'out'}(@_);
    } else {
        $body = $self->{'out'};
    }
    if ($body) {
        $user->SendMessage($body);
        $self->Activate;
    }
}

sub Hear($$) {
    my ($self, $message) = @_;

    my @matches = ( );
    if (defined($self->{'in'})) {
        if (ref $self->{'in'} eq 'ARRAY') {
            for (my $i = 0; $i < scalar @{$self->{'in'}}; $i++) {
                if ($message =~ /$self->{'in'}[$i]/i) {
                    push(@matches, $i + 1);
                    for (my $j = 1; $j < scalar @+; $j++) {
                        push(@matches, eval("\$$j"));
                    }
                    last;
                }
            }
        } elsif (ref $self->{'in'} eq 'CODE') {
            @matches = $self->{'in'}(@_);
        } else {
            if ($message =~ /$self->{'in'}/i) {
                push(@matches, 1);
                for (my $i = 1; $i < scalar @+; $i++) {
                    push(@matches, eval("\$$i"));
                }
            }
        }
    }

    return @matches;
}

sub Reply($$) {
    my ($self, $user) = @_;
    my $body = $self->{'reply'};
    if (ref $body eq 'ARRAY') {
        $body = $self->Random($body);
    }
    $user->SendMessage($body);
}

sub Random {
    if (scalar @_ > 0) {
        if (ref $_[0] eq 'Orangutan::Context') {
            shift;
        }
        if (scalar @_ > 0) {
            my $array = \@_;
            if (ref $_[0] eq 'ARRAY') {
                $array = $_[0];
            }
            return $array->[int(rand(scalar @{$array}))];
        }
    }
    return undef;
}

1;
