package Orangutan::Logger;

use strict;

sub new {
    my $class = shift;

    my $self = { };

    my %args = ( );
    while ($#_ >= 0) {
        $args{lc(pop)} = pop;
    }

    bless $self, $class;

    foreach my $arg (keys %args) {
        if ($arg eq 'directory') {
            $self->{'directory'} = $args{$arg};
        } elsif ($arg eq 'stdout') {
            $self->{'stdout'} = $args{$arg};
            binmode(STDOUT, ':utf8');
        }
    }

    return $self;
}

sub Log($$;@) {
    my ($self, $message, @args) = @_;
    my ($sec, $min, $hour, $day, $mon, $year) = localtime(time);
    my $filename = sprintf("%s/%04d%02d.log", $self->{'directory'}, $year + 1900, $mon + 1);
    if (open(LOG, '>>:utf8', $filename)) {
        printf(LOG "%02d.%02d.%02d %02d:%02d:%02d ", $year % 100, $mon + 1, $day, $hour, $min, $sec);
        printf(LOG $message, @args);
        print(LOG "\n");
        if ($self->{'stdout'}) {
            printf(STDERR "%02d.%02d.%02d %02d:%02d:%02d ", $year % 100, $mon + 1, $day, $hour, $min, $sec);
            printf(STDERR $message, @args);
            print(STDERR "\n");
        }
        close(LOG);
    }
}

sub LogUser($$$) {
    my ($self, $user, $message) = @_;
    my $path = $self->{'directory'}.'/'.$user->GetJabberUserID;
    if (-d $path || mkdir($path)) {
        my ($sec, $min, $hour, $day, $mon, $year) = localtime(time);
        my $filename = sprintf("%04d%02d.log", $year + 1900, $mon + 1);
        if (open(LOG, '>>:utf8', "$path/$filename")) {
            if (ref $message eq 'Net::XMPP::Presence') {
                my $status = undef;
                my $type = $message->GetType;
                my $show = $message->GetShow;
                if ($type eq 'unavailable') {
                    $status = 'offline';
                } elsif ($type eq '') {
                    if (($show eq 'away') || ($show eq 'xa')) {
                        $status = 'away';
                    } else {
                        $status = 'online';
                    }
                }
                if ($status) {
                    printf(LOG "%02d.%02d.%02d %02d:%02d:%02d %s went %s\n",
                        $year % 100, $mon + 1, $day, $hour, $min, $sec, $user->GetJabberUserID, $status);
                    if ($self->{'stdout'}) {
                        printf("%02d.%02d.%02d %02d:%02d:%02d %s went %s\n",
                            $year % 100, $mon + 1, $day, $hour, $min, $sec, $user->GetJabberUserID, $status);
                    }
                }
            } else {
                my $username;
                if ((caller())[0] eq 'Orangutan::User') {
                    $username = $main::config->Get('jabber', 'name');
                } else {
                    $username = $user->GetJabberUserID;
                }
                $message =~ s!\\!\\\\!g;
                $message =~ s!\r!\\r!g;
                $message =~ s!\n!\\n!g;
                printf(LOG "%02d.%02d.%02d %02d:%02d:%02d %s: %s\n",
                    $year % 100, $mon + 1, $day, $hour, $min, $sec, $username, $message);
                if ($self->{'stdout'}) {
                    printf("%02d.%02d.%02d %02d:%02d:%02d %s: %s\n",
                        $year % 100, $mon + 1, $day, $hour, $min, $sec, $username, $message);
                }
            }
            close(LOG);
        }
    }
}

1;
