package Orangutan::Task;

use strict;

use Data::Dumper;

use base 'Exporter';

our @EXPORT = qw(TASK_TIMESET TASK_SUBMITTED);

use constant TASK_TIMESET   => 0x01;
use constant TASK_SUBMITTED => 0x02;

sub new {
    my $class = shift;

    my $self = { };
    bless $self, $class;

    if (ref $_[0] eq 'Orangutan::Task') {
        my $task = shift;
        foreach my $member (keys %{$task}) {
            if ($member ne 'id') {
                $self->{$member} = $task->{$member};
            }
        }
        if ($self->IsSet(TASK_SUBMITTED)) {
            $self->Unset(TASK_SUBMITTED);
        }
    } else {
        $self->{'flags'} = 0;

        my %args = ( );
        while ($#_ >= 0) {
            $args{lc(pop)} = pop;
        }

        foreach my $arg (keys %args) {
            if ($arg eq 'id') {
                $self->SetID($args{$arg});
            } elsif
               (($arg eq 'name') ||
                ($arg eq 'task')) {
                $self->SetName($args{$arg});
            } elsif ($arg eq 'start') {
                $self->SetStart($args{$arg});
            } elsif ($arg eq 'issue') {
                $self->SetIssue($args{$arg});
            } elsif ($arg eq 'project') {
                if (ref $args{$arg} eq 'HASH') {
                    $self->SetProject(%{$args{$arg}});
                } else {
                    $self->SetProject(
                        name => $args{$arg}
                    );
                }
            } elsif ($arg eq 'activity') {
                if (ref $args{$arg} eq 'HASH') {
                    $self->SetActivity(%{$args{$arg}});
                } else {
                    $self->SetActivity(
                        name => $args{$arg}
                    );
                }
            } elsif ($arg eq 'end') {
                $self->SetEnd($args{$arg});
            } else {
                $main::logger->Log("unsupported task property: %s", $arg);
            }
        }

        if (!$self->GetStart) {
            $self->SetStart;
        }
    }

    return $self;
}

sub Save($$) {
    my ($self, $xml) = @_;

    my %attrs = ( );
    if (defined($self->{'id'})) {
        $attrs{'id'} = $self->{'id'};
    }
    $attrs{'start'} = $self->{'start'}->{'date'};
    if (defined($self->{'end'})) {
        $attrs{'end'} = $self->{'end'}->{'date'};
    }
    $xml->startTag('task', %attrs);
    if (defined($self->{'name'})) {
        $xml->dataElement('name', $self->{'name'});
    }
    if ($self->{'project'}) {
        $xml->startTag('project', id => $self->{'projectid'});
        $xml->dataElement('name', $self->{'project'});
        if ($self->{'issue'}) {
            $xml->emptyTag('issue', id => $self->{'issue'});
        }
        if ($self->{'activity'}) {
            $xml->startTag('activity', id => $self->{'activityid'});
            $xml->characters($self->{'activity'});
            $xml->endTag('activity');
        }
        $xml->endTag('project');
    }
    if ($self->{'flags'} != 0) {
        $xml->startTag('flags');
        if ($self->{'flags'} & TASK_TIMESET) {
            $xml->emptyTag('timeset');
        }
        if ($self->{'flags'} & TASK_SUBMITTED) {
            $xml->emptyTag('submitted');
        }
        $xml->endTag('flags');
    }
    $xml->endTag('task');
}

sub IsSet($$) {
    my ($self, $flag) = @_;
    return $self->{'flags'} & $flag;
}

sub Set($$) {
    my ($self, $flag) = @_;
    $self->{'flags'} |= $flag;
    ${$self->{'changes'}}++;
}

sub SetName($$) {
    my ($self, $task) = @_;
    $self->{'name'} = $task;
    if ($self->IsSet(TASK_SUBMITTED)) {
        $self->Unset(TASK_SUBMITTED);
    }
    ${$self->{'changes'}}++;
}

sub SetStart($;$) {
    my ($self, $time) = @_;
    if (defined($time)) {
        $self->Set(TASK_TIMESET);
    } else {
        $time = new Orangutan::Date;
        $self->Unset(TASK_TIMESET);
    }
    $self->{'start'} = $time;
    if ($self->IsSet(TASK_SUBMITTED)) {
        $self->Unset(TASK_SUBMITTED);
    }
    ${$self->{'changes'}}++;
}

sub SetIssue($@) {
    my ($self, $issue) = @_;
    $self->{'issue'} = $issue;
    if ($self->IsSet(TASK_SUBMITTED)) {
        $self->Unset(TASK_SUBMITTED);
    }
    ${$self->{'changes'}}++;
}

sub SetProject($@) {
    my $self = shift;

    my %args = ( );
    while ($#_ >= 0) {
        $args{lc(pop)} = pop;
    }

    delete($self->{'project'});
    delete($self->{'projectid'});

    foreach my $arg (keys %args) {
        if (($arg eq 'name') ||
            ($arg eq 'project')) {
            $self->{'project'} = $args{$arg};
        } elsif
           (($arg eq 'id') ||
            ($arg eq 'projectid') ||
            ($arg eq 'project_id')) {
            $self->{'projectid'} = $args{$arg};
        }
    }
    if ($self->IsSet(TASK_SUBMITTED)) {
        $self->Unset(TASK_SUBMITTED);
    }
    ${$self->{'changes'}}++;
}

sub SetActivity($@) {
    my $self = shift;

    my %args = ( );
    while ($#_ >= 0) {
        $args{lc(pop)} = pop;
    }

    delete($self->{'activity'});
    delete($self->{'activityid'});

    foreach my $arg (keys %args) {
        if (($arg eq 'name') ||
            ($arg eq 'activity')) {
            $self->{'activity'} = $args{$arg};
        } elsif
           (($arg eq 'id') ||
            ($arg eq 'activityid') ||
            ($arg eq 'activity_id')) {
            $self->{'activityid'} = $args{$arg};
        }
    }
    if ($self->IsSet(TASK_SUBMITTED)) {
        $self->Unset(TASK_SUBMITTED);
    }
    ${$self->{'changes'}}++;
}

sub SetEnd($;$) {
    my ($self, $time) = @_;
    if (!defined($time)) {
        $time = new Orangutan::Date;
    }
    $self->{'end'} = $time;
    if ($self->IsSet(TASK_SUBMITTED)) {
        $self->Unset(TASK_SUBMITTED);
    }
    ${$self->{'changes'}}++;
}

sub SetID($$) {
    my ($self, $id) = @_;
    $self->{'id'} = $id;
    ${$self->{'changes'}}++;
}

sub Unset($$) {
    my ($self, $flag) = @_;
    $self->{'flags'} &= ~$flag;
    ${$self->{'changes'}}++;
}

sub GetID($) {
    my $self = shift;
    return $self->{'id'};
}

sub GetName($) {
    my $self = shift;
    return $self->{'name'};
}

sub GetStart($) {
    my $self = shift;
    return $self->{'start'};
}

sub GetIssue($) {
    my $self = shift;
    return $self->{'issue'};
}

sub GetProject($;$) {
    my ($self, $query) = @_;

    if (defined($query)) {
        if ($query eq 'id') {
            return $self->{'projectid'};
        } else {
            return $self->{'project'};
        }
    } else {
        return $self->{'project'};
    }
}

sub GetActivity($;$) {
    my ($self, $query) = @_;

    if (defined($query)) {
        if ($query eq 'id') {
            return $self->{'activityid'};
        } else {
            return $self->{'activity'};
        }
    } else {
        return $self->{'activity'};
    }
}

sub GetEnd($) {
    my $self = shift;
    return $self->{'end'};
}

sub GetDuration($) {
    my $self = shift;
    if (defined($self->{'end'})) {
        return $self->{'end'}->Date - $self->{'start'}->Date;
    } else {
        return undef;
    }
}

sub XmlInit {
    my ($self, $user) = @_;
    $self->{'xml'} = { };
    $self->{'xml'}->{'path'} = [ ];
    $self->{'xml'}->{'user'} = $user;
}

sub XmlStart {
    my ($self, $expat, $element, %attrs) = @_;
    push(@{$self->{'xml'}->{'path'}}, $element);
    my $path = join('::', @{$self->{'xml'}->{'path'}});
    if ($path eq 'task') {
        foreach my $name (keys %attrs) {
            if ($name eq 'id') {
                $self->{'id'} = $attrs{$name};
            } elsif ($name eq 'start') {
                $self->{'start'} = new Orangutan::Date($attrs{$name});
            } elsif ($name eq 'end') {
                $self->{'end'} = new Orangutan::Date($attrs{$name});
            } else {
                $main::logger->Log("unsupported xml attribute: $name");
            }
        }
    } elsif ($path eq 'task::project') {
        foreach my $name (keys %attrs) {
            if ($name eq 'id') {
                $self->{'projectid'} = $attrs{$name};
            } else {
                $main::logger->Log("unsupported xml attribute: $name");
            }
        }
    } elsif ($path eq 'task::project::issue') {
        foreach my $name (keys %attrs) {
            if ($name eq 'id') {
                $self->{'issue'} = $attrs{$name};
            } else {
                $main::logger->Log("unsupported xml attribute: $name");
            }
        }
    } elsif ($path eq 'task::project::activity') {
        foreach my $name (keys %attrs) {
            if ($name eq 'id') {
                $self->{'activityid'} = $attrs{$name};
            } else {
                $main::logger->Log("unsupported xml attribute: $name");
            }
        }
    } elsif ($path eq 'task::flags::timeset') {
        $self->{'flags'} |= TASK_TIMESET;
    } elsif ($path eq 'task::flags::submitted') {
        $self->{'flags'} |= TASK_SUBMITTED;
    }
}

sub XmlChar {
    my ($self, $expat, $data) = @_;
    $data =~ s/^\s+//;
    $data =~ s/\s+$//;
    if ($data ne '') {
        my $path = join('::', @{$self->{'xml'}->{'path'}});
        if ($path eq 'task::name') {
            $self->{'name'} = $data;
        } elsif ($path eq 'task::project::name') {
            if (defined($self->{'projectid'})) {
                $self->{'project'} = $data;
            }
        } elsif ($path eq 'task::project::activity') {
            if (defined($self->{'activityid'})) {
                $self->{'activity'} = $data;
            }
        }
    }
}

sub XmlEnd {
    my ($self, $expat, $element) = @_;
    pop(@{$self->{'xml'}->{'path'}});
    if (scalar @{$self->{'xml'}->{'path'}} == 0) {
        $self->XmlFinal($expat, $element);
    }
}

sub XmlFinal {
    my ($self, $expat, $element) = @_;
    my $user = $self->{'xml'}->{'user'};
    $expat->setHandlers(
        Start => sub { $user->XmlStart(@_); },
        Char  => sub { $user->XmlChar(@_); },
        End   => sub { $user->XmlEnd(@_); },
    );
    $user->XmlEnd($expat, $element);
    delete($self->{'xml'});
}

1;
